//Stegano TGA - A Polecatsoft program
//The program will hide the file in parameter 3 in the bitmap specified in parameter 1,
//storing the result in parameter 2.
//The file can be retrieved using Stegano TGA Retrieve

#include <string>
#include <iostream>
#include <fstream>

#include "main.h"
#include "XorEncryption.h"
#include "hide.h"
#include "TGAhide.h"

using namespace std;

int steganoTgaHide(int argc, char **argv, hideArgs *cachedArgs)
{
	//for() counter variable
	int i;

	//Check if enough parameters are given
	cout << "You are now running the Stegano TGA function of MLSS.\n" << endl;
	if(argc < 3)
	{
		showHideHelp();
		return 0;
	}

	hideArgs args;
	if(cachedArgs)
		args = *cachedArgs;
	else
		args = getHideArgs(argc, argv);
	if(args.error)
		return 0;

	//File names
	string hostfile = args.hostfile, outputfile = args.outputfile, guestfile = args.guestfile;
	bool scramble = args.scramble;
	encryptClass *encryption = args.encryption;
	//How many bits of each hostfile byte will be used to hide data
	int bitsperbyte = args.bitsperbyte;
	void (*transferByte)(ifstream &input, ofstream &output, unsigned char writeByte, encryptClass *encryption) = args.transferByte;

	//hostfile bytes needed for each guestfile byte
	int bytesperbyte = 8 / bitsperbyte;

	//Check if all necessary parameters are given
	if(hostfile.empty())
	{
		cout << "Missing the non-optional option -h!" << endl;
		return 0;
	}
	if(guestfile.empty())
	{
		cout << "Missing the non-optional option -g!" << endl;
		return 0;
	}
	if(outputfile.empty())
		outputfile = "output.tga";

	//retrieve file size information
	//no longer needed: int inputSize = getFileSize(hostfile);
	int hideSize = getFileSize(guestfile);

	//Open existing TGA
	ifstream tgaIn(hostfile.c_str(), ios::binary);
	if(tgaIn.fail())
	{
		cout << "Failed to open " << hostfile << endl;
		return 0;
	}

	//Open existing file that should be hidden
	ifstream hideIn(guestfile.c_str(), ios::binary);
	if(hideIn.fail())
	{
		cout << "Failed to open " << guestfile << endl;
		return 0;
	}

	//Check header of TGA file
	const int HEADERSIZE = 12;
	const char perfectHeader[HEADERSIZE] = { 0, 0, 2, 0, 0, 0, 0, 0, 0, 0, 0, 0 };
	for(i = 0; i < HEADERSIZE; ++i)
	{
		char temp;
		if((temp = static_cast<char>(tgaIn.get())) != perfectHeader[i])
		{
			cout << "The file " << hostfile << " can not be used,\n"
				<< "the header (" << static_cast<unsigned int>(temp) << ") differs from " << static_cast<unsigned int>(perfectHeader[i]) << " at offset " << i << "." << endl;
			return 0;
		}
	}

	//Read additional header data
	unsigned short width, height;
	char colorDepth;
	tgaIn.read(reinterpret_cast<char *>(&width), sizeof(short));
	tgaIn.read(reinterpret_cast<char *>(&height), sizeof(short));
	tgaIn.read(&colorDepth, sizeof(char));
	if(colorDepth < 16)
	{
		cout << "Color depth ought to be at least 16!" << endl;
		return 0;
	}

	//Descriptor byte
	unsigned char dByte;
	tgaIn.read(reinterpret_cast<char *>(&dByte), sizeof(char));
	bool error = false;
	switch(colorDepth)
	{
	case 16:
		if((dByte & 0x0F) != 0x01)
			error = true;
		break;
	case 24:
		if((dByte & 0x0F) != 0x00)
			error = true;
		break;
	case 32:
		if((dByte & 0x0F) != 0x08)
			error = true;
		break;
	default:
		error = true;
		break;
	}
	if(dByte & 0xF0)
		error = true;
	if(error)
	{
		cout << "Error in image descriptor byte!\nHostfile is not suitable!" << endl;
		return 0;
	}

	//Do stuff for 16 bit images
	if(colorDepth == 16)
	{
		if(bitsperbyte == 4)
		{
			cout << "Color depth 16 bit is incompatible with 4 bits per byte.\n"
				"Truncating bits per byte to 2." << endl;
			bitsperbyte = 2;
		}
		if(bitsperbyte == 2)
		{
			transferByte = transfer16Byte2;
			bytesperbyte = 4;
		}
		if(bitsperbyte == 1)
		{
			transferByte = transfer16Byte1;
			bytesperbyte = 6;
		}
	}

	//We need to save the filename (stringlength + 1), the filesize (4) and the actual data (outputSize)
	int neededBytes = bytesperbyte * (static_cast<int>(guestfile.length()) + hideSize + 4 + 1);
	int availibleBytes = width * height * (colorDepth >> 3);
	//Check if image is big enough
	if(neededBytes > availibleBytes)
	{
		cout << "Hostfile is too small by " << neededBytes - availibleBytes << "!" << endl;
		return 0;
	}

	cout << "Hostfile was checked and approved.\nBeginning to hide " << guestfile << " with " << hideSize << " bytes." << endl;

	//Open output TGA file
	ofstream tgaOut(outputfile.c_str(), ios::binary);

	//Write header data to output TGA file
	for(i = 0; i < HEADERSIZE; ++i)
		tgaOut.write(&perfectHeader[i], 1);
	tgaOut.write(reinterpret_cast<char *>(&width), sizeof(width));
	tgaOut.write(reinterpret_cast<char *>(&height), sizeof(height));
	tgaOut.write(&colorDepth, 1);
	tgaOut.write(reinterpret_cast<char *>(&dByte), 1);

	//Perform actual hiding operation
	hideData(tgaIn, tgaOut, hideIn, guestfile, scramble, encryption, bytesperbyte, hideSize,
		transferByte, availibleBytes);
	if(encryption)
		delete encryption;
	cout << "File was successfully hidden." << endl;
	return 0;
}
